-- Construction of zone automaton

module ZoneReach (
   zoneAutomatonInclusion, zoneAutomatonEquality,
   futurZoneAutomatonInclusion, futurZoneAutomatonEquality
) where

import Data.List
import TimedAutomaton
import ZoneAutomaton
import Debug.Trace


-- Given a way to compare zones and a way to get the futur (before or after transition firing) compute the
-- zone automaton corresponding to a timed automaton
zoneAutomatonFun :: (Zone -> Zone -> Bool) -> (Int -> ZState -> [Transition] -> [ZTransition]) -> (Int -> Automaton -> ZState) -> Automaton -> ZAutomaton
zoneAutomatonFun matchZone firableTransitions getInitialState a = zoneAutomatonFun' [initial] [] []
   where
      maxConst = maxConstant a
      initial = getInitialState maxConst a
      zoneAutomatonFun' [] states transitions = ZAutomaton { 
                                                   zinitial=initial, 
                                                   zgoals=filter (\s@(ZState (l,_)) -> l `elem` (goals a)) states,
                                                   zstates=states,
                                                   ztransitions=transitions
                                                }
      zoneAutomatonFun' (s@(ZState (l,z)):open) seen transitionSeen =
         if (filter (\(ZState (l',z')) -> l == l' && matchZone z z') seen) == [] then
            let
               newTransitions = firableTransitions maxConst s (filter (\t -> source t == l) (transitions a))
               transitionSeen' = transitionSeen `union` newTransitions
               newStates = map ztarget newTransitions
               open' = open `union` (newStates \\ seen)
            in
               zoneAutomatonFun' open' (s:seen) transitionSeen'
         else
            zoneAutomatonFun' open (s:seen) transitionSeen


firableTransitionsFun :: (Int -> Zone -> Zone) -> (Int -> Zone -> Zone) -> Int -> ZState -> [Transition] -> [ZTransition]
firableTransitionsFun before after maxConst s@(ZState (l,z)) ts = fTFF ts []
   where
      futur = before maxConst z
      fTFF [] zts = zts
      fTFF (t:ts) zts = 
         let
            beforeResets = intersectGuard maxConst futur (guard t)
            fired = after maxConst (reset maxConst beforeResets (resets t))
         in
            if not (isEmpty beforeResets) then
               fTFF ts ((ZTransition { zsource=s, zlabel=(label t), ztarget=(ZState (target t,fired))}):zts)
            else
               fTFF ts zts

specialId a b = id b

firableTransitionsFuturFirst :: Int -> ZState -> [Transition] -> [ZTransition]
firableTransitionsFuturFirst = firableTransitionsFun extend specialId

firableTransitionsFuturLast :: Int -> ZState -> [Transition] -> [ZTransition]
firableTransitionsFuturLast  = firableTransitionsFun specialId extend


getInitialStateFuturFirst :: Int -> Automaton -> ZState
getInitialStateFuturFirst _ a = ZState (l,z)
   where
      l = initial a
      z = Zone (DBM (cs,vs))
      cs = (Clock "0") : (clocks a)
      k = length cs
      line = replicate k (Val 0)
      vs = replicate k line 

getInitialStateFuturLast :: Int -> Automaton -> ZState
getInitialStateFuturLast maxConst a = ZState (l,z)
   where
      ZState (l,z') = getInitialStateFuturFirst maxConst a
      z = extend maxConst z'

zoneAutomatonEquality = zoneAutomatonFun (==) firableTransitionsFuturFirst getInitialStateFuturFirst
futurZoneAutomatonEquality = zoneAutomatonFun (==) firableTransitionsFuturLast getInitialStateFuturLast
zoneAutomatonInclusion = zoneAutomatonFun include firableTransitionsFuturFirst getInitialStateFuturFirst
futurZoneAutomatonInclusion = zoneAutomatonFun include firableTransitionsFuturLast getInitialStateFuturLast